#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "iff.h"
#include "eor.h"

#define OBJT MKNAME('O', 'B', 'J', 'T')
#define ORNT MKNAME('O', 'R', 'N', 'T')
#define EXTE MKNAME('E', 'X', 'T', 'E')
#define APPR MKNAME('A', 'P', 'P', 'R')
#define INFO MKNAME('I', 'N', 'F', 'O')
#define POLY MKNAME('P', 'O', 'L', 'Y')
#define DETA MKNAME('D', 'E', 'T', 'A')
#define TRIS MKNAME('T', 'R', 'I', 'S')
#define FACE MKNAME('F', 'A', 'C', 'E')
#define MAPS MKNAME('M', 'A', 'P', 'S')
#define ATTR MKNAME('A', 'T', 'T', 'R')
#define TXMS MKNAME('T', 'X', 'M', 'S')
#define TXMD MKNAME('T', 'X', 'M', 'D')
#define TXMA MKNAME('T', 'X', 'M', 'A')
#define TXMP MKNAME('T', 'X', 'M', 'P')
#define VERT MKNAME('V', 'E', 'R', 'T')
#define REAL MKNAME('R', 'E', 'A', 'L')
#define LVL0 MKNAME('L', 'V', 'L', '0')
#define QUAD MKNAME('Q', 'U', 'A', 'D')
#define LINE MKNAME('L', 'I', 'N', 'E')
#define GRUP MKNAME('G', 'R', 'U', 'P')

static void FORM_REAL_APPR(struct IFFTag *, struct RealTag *);
static void FORM_REAL_APPR_POLY(struct IFFTag *, struct RealTag *);
static void FORM_REAL_APPR_POLY_TXMS(struct IFFTag *, struct RealTag *);
static void FORM_REAL_APPR_POLY_TRIS(struct IFFTag *, struct RealTag *);
static void FORM_REAL_APPR_POLY_LINE(struct IFFTag *, struct RealTag *);
static void FORM_REAL_APPR_POLY_QUAD(struct IFFTag *, struct RealTag *);

struct RealTag * eor_init(struct IFFStreamTag * stream)
{
  struct RealTag * Real;
  struct IFFTag * real = iff_init(stream);
  iff_name name = iff_readname(real);

  IFFCHK(name != REAL, real);
  real->name = name;
  real->size -= 4;
  Real = malloc(REALTAGSZ);
  IFFCHK(!Real, real);

  memset(Real, 0, REALTAGSZ);

  while (real->size)
  {
    iff_name name = iff_readname(real);
    iff_size size = iff_readsize(real);

    switch (name)
    {
      case INFO:
        iff_readdata(real, Real->name, 8);
        Real->name[8] = 0;
        break;

      case FORM:
      {
        iff_name name = iff_readname(real);
        struct IFFTag * form = iff_child(real, size - 4, name);

        switch (name)
        {
          case OBJT:
            iff_skip(form, size - 4);
            iff_align(form);
            break;

          case APPR:
            FORM_REAL_APPR(form, Real);
            break;

          default:
            UCHUNK(name, form);
            break;
        }
      } break;

      default:
      	UCHUNK(name, real);
        break;
    }
    real->size -= size + 8;
    IFFCHK(real->size < 0, real);
  }
  iff_close(real);
  return Real;
}

#define IFF_PROLOG(chunk) \
  while (chunk->size) \
  { \
    iff_name name = iff_readname(chunk); \
    iff_size size = iff_readsize(chunk); \
    switch (name) \
    { 

#define IFF_EPILOG(chunk) \
      default:            \
        UCHUNK(name, chunk); \
        break; \
    } \
    chunk->size -= size + 8; \
    IFFCHK(chunk->size < 0, chunk); \
  }   

#define IFF_SKIP(chunk, size) iff_skip(chunk, size); iff_align(chunk) 

static void FORM_REAL_APPR(struct IFFTag * appr, struct RealTag * Real)
{
  IFF_PROLOG(appr)

  case FORM:
  {
    iff_name name = iff_readname(appr);
    struct IFFTag * form = iff_child(appr, size - 4, name);
    switch (name)
    {
      case POLY: 
        FORM_REAL_APPR_POLY(form, Real);
        break;

      default:
        UCHUNK(name, appr);
        break;
    } 
  } break;

  IFF_EPILOG(appr)
}

static void FORM_REAL_APPR_POLY(struct IFFTag * poly, struct RealTag * Real)
{
  IFF_PROLOG(poly)

  case INFO:
  case ATTR:
     IFF_SKIP(poly, size);
     break;

  case VERT:
    Real->VertexCount = size / VERTEXTAGSZ;
    Real->VertexList = malloc(size);
    IFFCHK(!Real->VertexList, poly);
    iff_readdata(poly, Real->VertexList, size);
    break;

  case FORM:
  {
    iff_name name = iff_readname(poly);
    struct IFFTag * form = iff_child(poly, size - 4, name);
    switch (name)
    {
      case DETA:
        IFF_SKIP(poly, size - 4);
        break;

      case TXMS:
        FORM_REAL_APPR_POLY_TXMS(form, Real);
        break;

      case LINE:
        IFFCHK(Real->TrisFaceCount, form);
        IFFCHK(Real->QuadFaceCount, form);
        FORM_REAL_APPR_POLY_LINE(form, Real);
        break;

      case TRIS:
        IFFCHK(Real->QuadFaceCount, form);
        FORM_REAL_APPR_POLY_TRIS(form, Real);
        break;

      case QUAD: 
        FORM_REAL_APPR_POLY_QUAD(form, Real);
        break;

      case GRUP:
        IFF_SKIP(poly, size -4);
        break;

      default:
        UCHUNK(name, poly);
        break;
    }
  } break;

  IFF_EPILOG(poly)
}

static void FORM_REAL_APPR_POLY_TXMS(struct IFFTag * txms,
                                     struct RealTag * Real)
{
  iff_half AnimatedCount;
  iff_half StillCount;
  iff_sword TextureIndex = 0;
  IFF_PROLOG(txms)

  case TXMD:
    IFF_SKIP(txms, size);
    break;

  case INFO:
    StillCount = iff_readhalf(txms);
    AnimatedCount = iff_readhalf(txms);
    Real->TextureCount = StillCount + AnimatedCount;
    Real->TextureList = malloc(TEXTURETAGSZ * Real->TextureCount);

    IFFCHK(!Real->TextureList, txms);
    break;

  case TXMP:
  {
    iff_half xres, yres;
    iff_word size;

    iff_readdata(txms, Real->TextureList[TextureIndex].name, 8);
    Real->TextureList[TextureIndex].name[8] = 0;

    xres = iff_readhalf(txms);
    yres = iff_readhalf(txms);
    size = xres * yres;

    Real->TextureList[TextureIndex].xres = xres;
    Real->TextureList[TextureIndex].yres = yres;
    Real->TextureList[TextureIndex].bitmap = malloc(size);

    IFFCHK(!Real->TextureList[TextureIndex].bitmap, txms);

    iff_readdata(txms, Real->TextureList[TextureIndex].bitmap, size);
    iff_align(txms);

    TextureIndex++;
    IFFCHK(TextureIndex > Real->TextureCount, txms);
  } break;

  case TXMA:
  {
    iff_half xres, yres, frames;
    iff_word sres;

    iff_readdata(txms, Real->TextureList[TextureIndex].name, 8);
    Real->TextureList[TextureIndex].name[8] = 0;

    xres = iff_readhalf(txms);
    yres = iff_readhalf(txms);
    frames = iff_readhalf(txms);
    sres = xres * yres;

    Real->TextureList[TextureIndex].xres = xres;
    Real->TextureList[TextureIndex].yres = yres;
    Real->TextureList[TextureIndex].bitmap = malloc(sres);

    IFFCHK(!Real->TextureList[TextureIndex].bitmap, txms);

    iff_readdata(txms, Real->TextureList[TextureIndex].bitmap, sres);
    iff_skip(txms, size - sres - 14);
    iff_align(txms);

    TextureIndex++;
    IFFCHK(TextureIndex > Real->TextureCount, txms);
  } break;

  IFF_EPILOG(txms)
}

static void FORM_REAL_APPR_POLY_LINE(struct IFFTag * line,
                                     struct RealTag * Real)
{
  IFF_PROLOG(line)

  case FACE: 
    Real->LineFaceCount = size / LINEFACETAGSZ;
    Real->LineFaceList = malloc(size);
    IFFCHK(!Real->LineFaceList, line);
    iff_readdata(line, Real->LineFaceList, size);
    break;

  IFF_EPILOG(line)
}

static void FORM_REAL_APPR_POLY_TRIS(struct IFFTag * tris,
                                     struct RealTag * Real)
{
  IFF_PROLOG(tris)

  case FACE: 
    Real->TrisFaceCount = size / TRISFACETAGSZ;
    Real->TrisFaceList = malloc(size);
    IFFCHK(!Real->TrisFaceList, tris);
    iff_readdata(tris, Real->TrisFaceList, size);
    break;

  case MAPS: 
    Real->TrisMapCount = size / TRISMAPTAGSZ;
    Real->TrisMapList = malloc(size);
    IFFCHK(!Real->TrisMapList, tris);
    iff_readdata(tris, Real->TrisMapList, size);
    break;

  IFF_EPILOG(tris)
}

static void FORM_REAL_APPR_POLY_QUAD(struct IFFTag * quad,
                                     struct RealTag * Real)
{
  IFF_PROLOG(quad)

  case FACE: 
    Real->QuadFaceCount = size / QUADFACETAGSZ;
    Real->QuadFaceList = malloc(size);
    IFFCHK(!Real->QuadFaceList, quad);
    iff_readdata(quad, Real->QuadFaceList, size);
    break;

  case MAPS: 
    Real->QuadMapCount = size / QUADMAPTAGSZ;
    Real->QuadMapList = malloc(size);
    IFFCHK(!Real->QuadMapList, quad);
    iff_readdata(quad, Real->QuadMapList, size);
    break;

  IFF_EPILOG(quad)
}

#ifdef _TEST
int main(void)
{
  eor_init("lex.eor");
  return 0;
}
#endif
