#ifndef __qtr_included
#define __qtr_included

#include "config.h"

#include <math.h>
#include <string.h>
#include "cpppro.h"

#include "vec.h"
#include "mtx.h"

#define EPSILON 1e-5

#define QTRSIZE 4
#define QTRX 0
#define QTRY 1
#define QTRZ 2
#define QTRW 3

#define QTRCPY(l, r) VECCPY((l), (r)); (l)[3] = (r)[3]
#define QTRADD(l, r) VECADD((l), (r)); (l)[3] += (r)[3]
#define QTRSUB(l, r) VECSUB((l), (r)); (l)[3] -= (r)[3]
#define QTRDOT(l, r) (VECDOT((l), (r)) + (l)[3] * (r)[3])
#define QTRSCL(q, s) VECSCL((q), (s)); (q)[3] *= (s)
#define QTRMAG(q) (VECMAG((q)) + (q)[3] * (q)[3])
#define QTRCNJ(q) VECNEG((q))
#define QTRNEG(q) VECNEG((q)); (q)[3] = -(q)[3]
#define QTRNRM(q) (QTRDOT((q), (q)))
#define QTRNIL(q) VECNIL((q)); (q)[3] = 0.f
#define QTRIDN(q) VECNIL((q)); (q)[3] = 1.f
#define QTRINV(q) QTRCNJ((q)); { float m = 1.f / QTRNRM((q)); QTRSCL((q), m); }
#define QTRNIL(q) VECNIL((q)); (q)[3] = 0.f
#define QTRPAK(q, x, y, z, w) VECPAK((q), (x), (y), (z)); (q)[3] = (float) (w)
#define QTRUNP(q) VECUNP((q)), (q)[3]
#define QTRVEC(q, v) VECCPY((q), (v)); (q)[3] = 0.f
#define QTRPRT(q) \
printf("quaternion(" #q "): x:%9f y:%9f z:%9f w:%9f\n", QTRUNP(q))
 
void qtr_qtrmul(float *, float *, float *);
void qtr_vecmul(float *, float *, float, float, float);
void qtr_rotation(float *, float, float, float, float);
void qtr_mtx(float *, float *);
void qtr_slerp(float *, float *, float *, float);

#include "cppepi.h"

#endif /* __qtr_included */
